#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scrape each certified candidate's committees from the CAL-ACCESS site.
"""
import re
from time import sleep
from calaccess_scraped.management.commands import ScrapePageCommand
from calaccess_scraped.models import (
    Candidate,
    CandidateCommittee,
)


class Command(ScrapePageCommand):
    """
    Scrape each candidate's committees from the CAL-ACCESS site.
    """
    help = "Scrape each candidate's committees from the CAL-ACCESS site."

    def flush(self):
        """
        Delete records form related database tables.
        """
        CandidateCommittee.objects.all().delete()

    def scrape(self):
        """
        Make requests and parse markup into structured data.
        """
        self.header("Scraping candidate committees")

        # Set of unique scraped candidate ids
        candidate_ids = set(Candidate.objects.values_list(
            'scraped_id', flat=True))

        results = []

        for candidate_id in candidate_ids:
            url = '/Campaign/Candidates/Detail.aspx?id={}'.format(
                candidate_id)
            data = {}
            data['candidate_id'] = candidate_id
            data['committees'] = self.scrape_candidate_page(url)
            results.append(data)
            sleep(0.5)

        return results

    def scrape_candidate_page(self, url):
        """
        Pull the committees from a CAL-ACCESS candidate page.
        """
        soup = self.get_html(url)

        committees = []

        for table in soup.findAll('table')[2].findAll('table'):
            # Committees with a link
            c = table.find('a', {'class': 'sublink6'})
            if c:
                committees.append({
                    'name': c.text.strip(),
                    'scraped_id': re.match(r'.+id=(\d+)', c['href']).group(1).strip(),
                    'status': table.findAll('tr')[1].findAll('td')[1].text.strip()
                })
            # Committees with an ID but no link
            else:
                regex = re.compile(r'(.*)\(ID# (\d*)\)')
                c = table.find('span', string=regex)
                if c:
                    matches = re.match(regex, c.text)
                    committees.append({
                        'name': matches.group(1).strip(),
                        'scraped_id': matches.group(2).strip(),
                        'status': table.findAll('tr')[1].findAll('td')[1].text.strip()
                    })

        return committees

    def save(self, results):
        """
        Save results of scrape to related database tables.
        """
        self.log('Processing %s committees.' % len(results))

        for result in results:
            for committee_data in result['committees']:
                # Add the candidate id to the committee data
                committee_data['candidate_id'] = result['candidate_id']
                committee_obj, c = CandidateCommittee.objects \
                    .get_or_create(
                        **committee_data
                    )
                if c and self.verbosity > 2:
                    self.log('Created %s' % committee_obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Run all scraper commands.
"""
from django.core.management import call_command
from calaccess_scraped.management.commands import ScrapeCommand


class Command(ScrapeCommand):
    """
    Run all scraper commands.
    """
    help = "Run all scraper commands."

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        call_command(
            'scrapecalaccesspropositions',
            verbosity=self.verbosity,
            no_color=self.no_color,
            force_flush=self.force_flush,
            force_download=self.force_download,
            update_cache=self.update_cache,
        )

        call_command(
            'scrapecalaccesscandidates',
            verbosity=self.verbosity,
            no_color=self.no_color,
            force_flush=self.force_flush,
            force_download=self.force_download,
            update_cache=self.update_cache,
        )

        call_command(
            'scrapecalaccessincumbents',
            verbosity=self.verbosity,
            no_color=self.no_color,
            force_flush=self.force_flush,
            force_download=self.force_download,
            update_cache=self.update_cache,
        )
        # Should we be running scrapecalaccesscandidatecommittees too?



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scrape list of certified candidates from the CAL-ACCESS site.
"""
import re
from time import sleep
from six.moves.urllib.parse import urljoin
from calaccess_scraped.management.commands import ScrapePageCommand
from calaccess_scraped.models import (
    Candidate,
    CandidateElection
)


class Command(ScrapePageCommand):
    """
    Scrape list of certified candidates for each election on CAL-ACCESS site.
    """
    help = "Scrape certified candidates for each election on the CAL-ACCESS \
    site."

    def flush(self):
        """
        Delete records form related database tables.
        """
        Candidate.objects.all().delete()
        CandidateElection.objects.all().delete()

    def scrape(self):
        """
        Make requests and parse markup into structured data.
        """
        self.header("Scraping election candidates")

        soup = self.get_html(
            '/Campaign/Candidates/list.aspx?view=certified&electNav=93'
        )

        # Get all the links out
        links = soup.findAll('a', href=re.compile(r'^.*&electNav=\d+'))

        # Drop the link that says "prior elections" because it's a duplicate
        links = [
            l for l in links
            if l.find_next_sibling('span').text != 'Prior Elections'
        ]

        # Loop through the links
        results = {}
        for i, link in enumerate(links):
            # Get each page and its data
            url = urljoin(self.base_url, link["href"])
            data = self.scrape_election_page(url)
            # Add the name of the election
            data['name'] = link.find_next_sibling('span').text.strip()
            # The index value is used to preserve sorting of elections,
            # since multiple elections may occur in a year.
            # BeautifulSoup goes from top to bottom,
            # but the top most election is the most recent so it should
            # have the highest id.
            data['sort_index'] = len(links) - i
            # Add it to the list
            results[url] = data
            # Take a rest
            sleep(0.5)

        return results

    def scrape_election_page(self, url):
        """
        Pull the elections and candidates from a CAL-ACCESS page.
        """
        # Go and get the page
        soup = self.get_html(url)

        races = {}
        # Loop through all the election sets on the page
        for section in soup.findAll('a', {'name': re.compile(r'[a-z]+')}):

            # Check that this data matches the structure we expect.
            section_name_el = section.find('span', {'class': 'hdr14'})

            # If it doesn't, skip this one
            if not section_name_el:
                continue

            # Loop through all the rows in the section table
            for office in section.findAll('td'):

                # Check that this data matches the structure we expect.
                title_el = office.find('span', {'class': 'hdr13'})

                # If it doesn't, skip
                if not title_el:
                    continue

                office_name = title_el.text

                # Log what we're up to
                if self.verbosity > 2:
                    self.log('Scraping office %s' % office_name)

                # Pull the candidates out
                candidates = []
                for c in office.findAll('a', {'class': 'sublink2'}):
                    candidates.append({
                        'name': c.text,
                        'scraped_id': re.match(
                            r'.+id=(\d+)',
                            c['href']
                        ).group(1),
                        'url': urljoin(self.base_url, url),
                    })

                for c in office.findAll('span', {'class': 'txt7'}):
                    candidates.append({
                        'name': c.text,
                        'scraped_id': '',
                        'url': urljoin(self.base_url, url),
                    })

                # Add it to the data dictionary
                races[office_name] = candidates

        return {
            'scraped_id': int(re.match(r'.+electNav=(\d+)', url).group(1)),
            'races': races,
            'url': urljoin(self.base_url, url)
        }

    def save(self, results):
        """
        Save results of scrape to related database tables.
        """
        self.log('Processing %s elections.' % len(results))

        # Loop through all the results
        for url, election_data in results.items():

            self.log('Processing %s' % election_data['name'])
            election_obj, c = CandidateElection.objects.get_or_create(
                name=election_data['name'].strip(),
                scraped_id=election_data['scraped_id'],
                sort_index=election_data['sort_index'],
                url=url,
            )

            if c and self.verbosity > 2:
                self.log('Created %s' % election_obj)

            # Loop through each of the races
            for office_name, candidates in election_data['races'].items():

                # Loop through each of the candidates
                for candidate_data in candidates:

                    # Create the candidate object
                    candidate_obj, c = Candidate.objects.get_or_create(
                        name=candidate_data['name'].strip(),
                        scraped_id=candidate_data['scraped_id'],
                        office_name=office_name.strip(),
                        url=candidate_data['url'],
                        election=election_obj,
                    )

                    if c and self.verbosity > 2:
                        self.log('Created %s' % candidate_obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base classes for custom management commands.
"""
import os
import re
import logging
import requests
from bs4 import BeautifulSoup
from django.conf import settings
from django.utils import timezone
from six.moves.urllib.parse import urljoin
from django.utils.termcolors import colorize
from six.moves.urllib.request import url2pathname
from django.core.management.base import BaseCommand
from calaccess_scraped.models.tracking import ScrapedDataVersion
from calaccess_scraped.decorators import retry
logger = logging.getLogger(__name__)


class CalAccessCommand(BaseCommand):
    """
    Base class for all custom CalAccess-related management commands.
    """
    def handle(self, *args, **options):
        """
        Sets options common to all commands.

        Any command subclassing this object should implement its own
        handle method, as is standard in Django, and run this method
        via a super call to inherit its functionality.
        """
        # Set global options
        self.verbosity = options.get("verbosity")
        self.no_color = options.get("no_color")

        # Start the clock
        self.start_datetime = timezone.now()

    def header(self, string):
        """
        Writes out a string to stdout formatted to look like a header.
        """
        logger.debug(string)
        if not getattr(self, 'no_color', None):
            string = colorize(string, fg="cyan", opts=("bold",))
        self.stdout.write(string)

    def log(self, string):
        """
        Writes out a string to stdout formatted to look like a standard line.
        """
        logger.debug(string)
        if not getattr(self, 'no_color', None):
            string = colorize("%s" % string, fg="white")
        self.stdout.write(string)

    def success(self, string):
        """
        Writes out a string to stdout formatted green to communicate success.
        """
        logger.debug(string)
        if not getattr(self, 'no_color', None):
            string = colorize(string, fg="green")
        self.stdout.write(string)

    def warn(self, string):
        """
        Writes string to stdout formatted yellow to communicate a warning.
        """
        logger.warn(string)
        if not getattr(self, 'no_color', None):
            string = colorize(string, fg="yellow")
        self.stdout.write(string)

    def failure(self, string):
        """
        Writes string to stdout formatted red to communicate failure.
        """
        logger.error(string)
        if not getattr(self, 'no_color', None):
            string = colorize(string, fg="red")
        self.stdout.write(string)

    def duration(self):
        """
        Calculates how long command has been running and writes it to stdout.
        """
        duration = timezone.now() - self.start_datetime
        self.stdout.write('Duration: {}'.format(str(duration)))
        logger.debug('Duration: {}'.format(str(duration)))

    def __str__(self):
        return re.sub(r'(.+\.)*', '', self.__class__.__module__)


class ScrapeCommand(CalAccessCommand):
    """
    Base management command for scraping the CAL-ACCESS website.
    """
    base_url = 'http://cal-access.sos.ca.gov/'
    cache_dir = os.path.join(
        settings.BASE_DIR,
        ".scraper_cache"
    )

    def add_arguments(self, parser):
        """
        Adds custom arguments specific to this command.
        """
        parser.add_argument(
            '--flush',
            action='store_true',
            dest='force_flush',
            default=False,
            help='Flush database tables',
        )
        parser.add_argument(
            '--force-download',
            action='store_true',
            dest='force_download',
            default=False,
            help='Force the scraper to download URLs even if they are cached',
        )
        parser.add_argument(
            '--cache-only',
            action='store_false',
            dest='update_cache',
            default=True,
            help="Skip the scraper's update checks. Use only cached files.",
        )

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(ScrapeCommand, self).handle(*args, **options)

        self.force_flush = options.get("force_flush")
        self.force_download = options.get("force_download")
        self.update_cache = options.get("update_cache")

        os.path.exists(self.cache_dir) or os.mkdir(self.cache_dir)

    def get_scraped_version(self):
        """
        Get or create the current processed version.

        Return a tuple (ProcessedDataVersion object, created), where
        created is a boolean specifying whether a version was created.
        """
        return ScrapedDataVersion.objects.create()


class ScrapePageCommand(ScrapeCommand):
    """
    Base management command for scraping a page/subsection of the CAL-ACCESS website.
    """
    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(ScrapePageCommand, self).handle(*args, **options)

        os.path.exists(self.cache_dir) or os.mkdir(self.cache_dir)

        if self.force_flush:
            self.flush()
        results = self.scrape()
        self.save(results)

    @retry(requests.exceptions.RequestException)
    def get_url(self, url, retries=1, request_type='GET'):
        """
        Returns the response from a URL, retries if it fails.
        """
        headers = {
            'User-Agent': 'California Civic Data Coalition \
            (cacivicdata@gmail.com)',
        }
        if self.verbosity > 2:
            self.log(" Making a {} request for {}".format(request_type, url))
        return getattr(requests, request_type.lower())(url, headers=headers)

    def get_headers(self, url):
        """
        Returns a dict with metadata about the current CAL-ACCESS snapshot.
        """
        response = self.get_url(url, request_type='HEAD')
        length = int(response.headers['content-length'])
        return {
            'content-length': length,
        }

    def get_html(self, url, retries=1, base_url=None):
        """
        Makes request for a URL and returns HTML as a BeautifulSoup object.
        """
        # Put together the full URL
        full_url = urljoin(base_url or self.base_url, url)
        if self.verbosity > 2:
            self.log(" Retrieving data for {}".format(url))

        # Pull a cached version of the file, if it exists
        cache_path = os.path.join(
            self.cache_dir,
            url2pathname(url.strip("/"))
        )
        if os.path.exists(cache_path) and not self.force_download:
            # Make a HEAD request for the file size of the live page
            if self.update_cache:
                cache_file_size = os.path.getsize(cache_path)
                head = self.get_headers(full_url)
                web_file_size = head['content-length']

                if self.verbosity > 2:
                    msg = " Cached file sized {}. Web file size {}."
                    self.log(msg.format(
                        cache_file_size,
                        web_file_size
                    ))

            # If our cache is the same size as the live page, return the cache
            if not self.update_cache or cache_file_size == web_file_size:
                if self.verbosity > 2:
                    self.log(" Returning cached {}".format(cache_path))
                html = open(cache_path, 'r').read()
                return BeautifulSoup(html, "html.parser")

        # Otherwise, retrieve the full page and cache it
        try:
            response = self.get_url(full_url)
        except requests.exceptions.HTTPError as e:
            # If web requests fails, fall back to cached file, if it exists
            if os.path.exists(cache_path):
                if self.verbosity > 2:
                    self.log(" Returning cached {}".format(cache_path))
                html = open(cache_path, 'r').read()
                return BeautifulSoup(html, "html.parser")
            else:
                raise e

        # Grab the HTML and cache it
        html = response.text
        if self.verbosity > 2:
            self.log(" Writing to cache {}".format(cache_path))
        cache_subdir = os.path.dirname(cache_path)
        os.path.exists(cache_subdir) or os.makedirs(cache_subdir)
        with open(cache_path, 'w') as f:
            f.write(html)

        # Finally return the HTML ready to parse with BeautifulSoup
        return BeautifulSoup(html, "html.parser")

    def flush(self):
        """
        This method should empty out database tables filled by this command.
        """
        raise NotImplementedError

    def scrape(self):
        """
        This method should perform the actual scraping.

        Returns the structured data.
        """
        raise NotImplementedError

    def save(self, results):
        """
        This method should process structured data returned by `build_results`.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scrape links between filers and propositions from the CAL-ACCESS site.
"""
import re
from time import sleep
from six.moves.urllib.parse import urljoin
from calaccess_scraped.management.commands import ScrapePageCommand
from calaccess_scraped.models import (
    PropositionElection,
    Proposition,
    PropositionCommittee
)


class Command(ScrapePageCommand):
    """
    Scrape links between filers and propositions from official CAL-ACCESS site.
    """
    help = "Scrape links between filers and propositions from the official \
    CAL-ACCESS site."

    def flush(self):
        """
        Delete records form related database tables.
        """
        PropositionCommittee.objects.all().delete()
        Proposition.objects.all().delete()
        PropositionElection.objects.all().delete()

    def scrape(self):
        """
        Make requests and parse markup into structured data.
        """
        self.header("Scraping propositions")

        # Build the link list from the 2013 page because otherwise the
        # other years are hidden under the "Historical" link.
        soup = self.get_html('Campaign/Measures/list.aspx?session=2013')

        # Filter links for uniqueness.
        links = soup.findAll('a', href=re.compile(r'^.*\?session=\d+'))
        links = list(set([link['href'] for link in links]))

        results = {}
        for link in links:
            results[urljoin(self.base_url, link)] = self.scrape_year_page(link)

        # Pass it out
        return results

    def scrape_year_page(self, url):
        """
        Scrape data page with list of props in a particular election year.
        """
        # Get the URL of the year page
        soup = self.get_html(url)

        # Loop through all the tables on the page
        data_dict = {}
        table_list = soup.findAll(
            'table',
            {'id': re.compile(r'ListElections1__[a-z0-9]+')}
        )
        for table in table_list:

            # Pull the title
            election_name = table.select('caption span')[0].text

            # Get a list of the propositions in this table
            prop_links = table.findAll('a')

            # Log what we're up to
            if self.verbosity > 2:
                msg = " Scraped: %s (%s props)"
                msg = msg % (
                    election_name,
                    len(prop_links),
                )
                self.log(msg)

            # Scrape them one by one
            prop_list = [
                self.scrape_prop_page('/Campaign/Measures/%s' % link['href'])
                for link in prop_links
            ]

            # Add the data to our data dict
            data_dict[election_name] = prop_list

        # Take a rest
        sleep(0.5)

        # Pass the data back out
        return data_dict

    def scrape_prop_page(self, url):
        """
        Scrape data from a proposition detail page.
        """
        # Pull the page
        soup = self.get_html(url)

        # Create a data dictionary to put the good stuff in
        data_dict = {'url': urljoin(self.base_url, url)}

        # Add the title and id out of the page
        data_dict['name'] = soup.find('span', id='measureName').text

        data_dict['id'] = re.match(r'.+id=(\d+)', url).group(1)

        data_dict['committees'] = []
        # Loop through all the tables on the page
        # which contain the committees on each side of the measure
        for table in soup.findAll('table', cellpadding='4'):

            # Pull the data box
            data = table.findAll('span', {'class': 'txt7'})

            # The URL
            committee_url = table.find('a', {'class': 'sublink2'})

            # The name
            name = committee_url.text

            # ID sometimes refers to xref_filer_id rather than filer_id_raw
            id_ = data[0].text

            # Does the committee support or oppose the measure?
            position = data[1].text.strip()

            # Put together a data dictionary and add it to the list
            data_dict['committees'].append({
                'name': name,
                'id': id_,
                'position': position,
                'url': urljoin(self.base_url, committee_url['href']),
            })

        if self.verbosity > 2:
            msg = " Scraped: %s (%s committees)"
            msg = msg % (
                data_dict['name'],
                len(data_dict['committees'])
            )
            self.log(msg)

        # Take a rest
        sleep(0.5)

        # Pass the data out
        return data_dict

    def save(self, results):
        """
        Save results of scrape to related database tables.
        """
        # For each year page
        for url, d in results.items():
            # For each election on that page
            for election_name, prop_list in d.items():
                # Get or create election object
                election_obj, c = PropositionElection.objects \
                    .get_or_create(
                        name=election_name.strip(),
                        url=url,
                    )
                # Loop through propositions
                for prop_data in prop_list:
                    # Get or create proposition object
                    prop_obj, c = Proposition.objects.get_or_create(
                        name=prop_data['name'].strip(),
                        scraped_id=prop_data['id'],
                        url=prop_data['url'],
                        election=election_obj
                    )
                    # Log it
                    if c and self.verbosity > 2:
                        self.log('Created %s' % prop_obj)

                    # Now loop through the committees
                    for committee in prop_data['committees']:
                        # Get or create it
                        committee_obj, c = PropositionCommittee \
                            .objects.get_or_create(
                                name=committee['name'].strip(),
                                scraped_id=committee['id'],
                                position=committee['position'],
                                url=committee['url'],
                                proposition=prop_obj,
                            )

                        # Log it
                        if c and self.verbosity > 2:
                            self.log('Created %s' % committee_obj)
